__doc__='''

Simple EMarket product implementation.

An EMarket is a collection of classes, folders and documents that implements
a simple online shopping area. There are two main folders that keep things
organized. 1) A shoppers folder, for all the online shoppers and 2) a marketItems
folder to keep the items to be sold.  The marketItems folder can have sub-folders
which act as 'categories' for different sorts of products.

Steve Spicklemire
Silicon Prairie Ventures Inc.
steve@spvi.com

'''

__version__='$Revision: 1.67 $'[11:-2]

from Globals import HTMLFile, MessageDialog, Persistent, DateTime
from Persistence import PersistentMapping
import OFS.SimpleItem, Acquisition, AccessControl.Role
import OFS.Folder, OFS.Document
import Products.MailHost.MailHost
import Shopper
import MarketItem
import traceback
import StringIO
import string
from socket import gethostname
import time
import random

Folder = OFS.Folder.Folder
Document = OFS.Document.Document

EMarketError = 'Emarket Error'

#
# these are kinda random
#

randWords = ['foo','bar','shoo','fork','spoon','table','car','stairs',
             'dog','cat','heart','sprint','flower','close','run','buy',
             'bell','glass','bike','floor','blink','hover','sun','moon']
randSymbs = ['%','$','#','*']

#
# Prepare a bunch of sources to include as default initial
# docs for the folders.....
#

def GetHTML(name):
    return HTMLFile(name, globals())

# --- Main EMarket Folder DTML Sources... -----

mainMarketDTML = {'index_html':(GetHTML('eMarketMain'), "Main EMarket Page"),
                  'newUserFormFields':(GetHTML('newUserFields'), "Input fields for new user"),
                  'setUserCookie':(GetHTML('setUserCookie'), "Prepare to set the cookie...."),
                  'register':(GetHTML('register'), "Register as a past user...."),
                  'sendPassword':(GetHTML('sendPassword'), "Send a forgotten password"),
                  'redirectToShoppingCart':(GetHTML('redirectToShoppingCart'), "Redirect to Shopping Cart Area"),
                  'redirectToSalesItems':(GetHTML('redirectToSalesItems'), "Redirect to Sales Items Area"),
                  'redirectToEMarket':(GetHTML('redirectToEMarket'), "Redirect to Main Market"),
                  'redirectToEditInfo':(GetHTML('redirectToEditInfo'), "Redirect to Edit Page"),
                  'redirectToPastTransactions':(GetHTML('redirectToPastTransactions'), "Redirect to Past Transactions"),
                  'addToShoppingCart':(GetHTML('addToShoppingCart'), "Add arbitrary item to shopping cart."),
                  'standard_html_footer':(GetHTML('footer'),"Footer dtml"),
                  'logOut':(GetHTML('logOut'),"log out of server..."),
                  'addShopperWrapper':(GetHTML('addShopperWrapper'),"Add New Shopper"),
                  'checkForValues':(GetHTML('checkForValues'),"Check for missing values"),
                  'productListing':(GetHTML('productListing'),"Generate a nice category list."),
                  'searchresults':(GetHTML('searchresults'),"Catalog Search Results"),
                  }

# --- Market Item folder DTML -----

marketItemDTML = {'index_html':(GetHTML('eMarketItemMain'), 'MarketItem Template Page'),
                  'addToCart': (GetHTML('addToCart'), 'Form To add MarketItems to a cart.'),
                  'shopperMarketItemAddForm': (GetHTML('shopperMarketItemAdd'), 'Form To add MarketItems to sell.'),
                  'shopperMarketItemEditForm': (GetHTML('shopperMarketItemEditForm'), 'Form To add MarketItems to sell.'),
                  'listActiveOffers': (GetHTML('listActiveOffers'), 'Active offers for sales item.'),
                  'acceptOffer': (GetHTML('acceptOffer'), 'Accept offer from a buyer.'),
                  'displayBids': (GetHTML('displayBids'), 'Display current bids on an item.'),
                  }

# --- Shopper Folder DTML ----

shopperDTML = {'addNewItem':(GetHTML('shopperAddItem'),'Response template for adding item'),
               'index_html':(GetHTML('shopperIndex'),'Shopper Template'),
               'showMyBasket':(GetHTML('showMyBasket'),'Shopping basket template'),
               'showMySales':(GetHTML('showMySales'),'Shopping basket template'),
               'processOrder':(GetHTML('processOrder'),"Process a shopper's order."),
               'shoppingCart':(GetHTML('shoppingCart'), "Shopping cart area"),
               'shopperSalesItems':(GetHTML('salesItems'), "Display Sales Items area"),
               'updateCart':(GetHTML('updateCart'), "Update the cart area"),
               'updateSales':(GetHTML('updateSales'), "Update the sales booth area"),
               'editSelfShopper':(GetHTML('editSelfShopper'),"Edit a Shopper's Personal Info"),
               'submitOrder':(GetHTML('submitOrder'),"Submit an order for processing"),
               'listValidatedTransactions':(GetHTML('listValidated'),
                                            "list of validated transactions"),
               'acceptableOffer':(GetHTML('acceptableOffer'), "You have made an acceptable offer..."),
               'confirmOffer':(GetHTML('confirmOffer'),"Confirm Your Offer..."),
               'confirmOfferForm':(GetHTML('confirmOfferForm'),"Confirm Your Offer..."),
               'pendingCart':(GetHTML('pendingCart'),"Confirm Your Offer..."),
               }

wampumboxDTML = {'index_html':(GetHTML('wampum_submitCCTrans'),"Form to submit CC purchase through Wampum"),
                 'processCCTrans':(GetHTML('wampum_processCCTrans'),"Submit CC purchase through Wampum"),
                }


def cmpFolders(folder1, folder2):
    if folder1.title < folder2.title:
        return -1
    elif folder1.title > folder2.title:
        return 1
    else:
        return 0


def delDocsFromFolder( theFolder, theDocs ):
    """ Delete the listed docs from the given folder. """
    errs = StringIO.StringIO()
    
    for key in theDocs.keys():
        try:
            theFolder._delObject(key)
        except:
            traceback.print_exc(file=errs)

    errs.seek(0)
    return errs.read()
    
def addDocsToFolder( theFolder, theDocs):
    """ add the listed docs to the given folder """
    
    errs = StringIO.StringIO()

    for key in theDocs.keys():
        id = key
        dtml,title = theDocs[key]
        try:
            thedtml = open(dtml.raw).read()
        except:
            raise RuntimeError, "Can't find " + dtml.raw

        try:
            theFolder.manage_addDocument(id, title=title, file=thedtml)
        except:
            traceback.print_exc(file=errs)

    errs.seek(0)
    return errs.read()

def updateDocsInFolder( theFolder, theDocs, REQUEST=None ):
    """ update the documents in a folder from theDocs collection """

    processing = []
    errs = StringIO.StringIO()
    
    for key in theDocs.keys():
        dtml, title = theDocs[key]
        theDTML = open(dtml.raw).read()
        
        if hasattr(theFolder, key):
            theDoc = getattr(theFolder, key)
            theDoc.manage_edit(data = theDTML, title=title, REQUEST=REQUEST)
            processing.append('Updated document ' + key )
        else:
            theFolder.manage_addDTMLMethod(key, title=title, file=theDTML, REQUEST=REQUEST)
            processing.append('Added document ' + key )

    return '<pre>' + string.join(processing,'<br>\n') + '</pre>'


try:
    defaulthost = gethostname()
except:
    defaulthost = 'localhost'


addMarketItemsFolderForm=HTMLFile('marketItemsFolderAdd', globals())

def addMarketItemsFolder(self, id, title='', REQUEST=None):
    """ method to add market items folders to a folder. """
    newMarketF = Folder()
    newMarketF.id = id
    newMarketF.title = title
    newMarketF.meta_types = ({'name':'MarketItem', 'action':'manage_addMarketItemForm'},
                {'name':'MarketItemsFolder', 'action':'manage_addMarketItemsFolderForm'})
    newMarketF.icon = 'misc_/EMarket/marketItemFolder'

    self._setObject(id, newMarketF)

    if REQUEST is None:
        result = newMarketF
    else:
        result = MessageDialog(
            title = "OK",
            message='<strong>%s</strong> Has beed added as a MarketItems Folder.' % id,
            action ='./manage_main',
            )

    return result

addEMarketForm=HTMLFile('eMarketAdd', globals(), defaulthost = defaulthost)
    
def addEMarket(self, id, title='', smtp_host = defaulthost, localhost = defaulthost, processURL='',
               localSecret='', contactInfo='', EMarketURL='',
               default_html = '', requireLogin = 1, addToCartText = 'Add To Cart', install_wampumbox = 1, REQUEST=None):

    """Add an EMarket to a folder

    The argument 'self' will be bound to a folder.

    """
    #
    
    newMainF = EMarket()
    newMainF.id = id
    newMainF.title = title
    newMainF.smtp_host = smtp_host
    newMainF.localhost = localhost
    if processURL=='' and install_wampumbox:
      processURL = 'https' + self.absolute_url()[4:] + '/' + id + \
        '/wampumbox/'
    newMainF.processURL = processURL
    newMainF.localSecret = localSecret
    newMainF.contactInfo = contactInfo
    newMainF.EMarketURL = EMarketURL
    newMainF.default_html = default_html
    newMainF.requireLogin = requireLogin
    newMainF.addToCartText = addToCartText
    newMainF.biddingAllowed = 0  # default to no...
    newMainF.defaultToAuction = 0 # have new market items default to auction sales
    newMainF.randWords = randWords

    self._setObject(id, newMainF)

    if id == 'shoppers' or id == 'marketItems' or id == 'trasactions':
        raise EMarketError, "The id %s is illegal." % id

    newMarketF = addMarketItemsFolder(newMainF, 'marketItems', "The Market Items")
    
    newShopperF = Folder()
    newShopperF.id = 'shoppers'
    newShopperF.title = "Market Shoppers"
    newShopperF.meta_types = ({'name':'Shopper', 'action':'manage_addShopperForm'},)
    newShopperF.icon = 'misc_/EMarket/shopperFolder'
    
    newMainF._setObject(newShopperF.id, newShopperF)

    if install_wampumbox:
        newWampumboxF = Folder()
        newWampumboxF.id = 'wampumbox'
        newWampumboxF.title = 'Wampum Transaction Processor'
        newWampumboxF.meta_types = ()
        newWampumboxF.icon = 'misc_/EMarket/wampumbox'
        Products.Wampum.WampumGenerator.addWampum(newWampumboxF,'Wampum','Wampum Generator',ccps_host='',cybercash_id='',merchant_key='')

        newMainF._setObject(newWampumboxF.id, newWampumboxF)

    newTransF = Folder()
    newTransF.id = 'transactions'
    newTransF.title = "Transactions"
    newTransF.icon = 'misc_/EMarket/marketItemFolder'
    
    newMainF._setObject(newTransF.id, newTransF)

    addDocsToFolder(newMainF, mainMarketDTML)
    addDocsToFolder(newMarketF, marketItemDTML)
    addDocsToFolder(newShopperF, shopperDTML)
    if install_wampumbox:
        addDocsToFolder(newWampumboxF, wampumboxDTML)
#        x=newWampumboxF.processCCTrans
#        x._proxy_roles=('Manager',)
#        newWampumboxF.processCCTrans=x

    newMainF.shopperFolder = newShopperF
    newMainF.marketFolder = newMarketF
    newMainF.transactionFolder = newTransF
    if install_wampumbox:
        newMainF.wampumbox = newWampumboxF
    else:
        newMainF.wampumbox = None
    
    Products.MailHost.MailHost.add(newMainF, 'eMarketMailhost',
                                   smtp_host = smtp_host,
                                   localhost = localhost)

    theMailHost = getattr(newMainF,'eMarketMailhost')  ### TODO... change documents that send mail to be proxies for Manager

    Products.ZCatalog.ZCatalog.manage_addZCatalog(newMainF,'Catalog',
                                                  'Market Items Catalog')
    newMainF.Catalog._catalog.addIndex('isAvailable','FieldIndex')

    if REQUEST is not None:
        return self.manage_main(self, REQUEST, update_menu=1)

class Dummy:

    def __init__(self, title=None):
        self.title = title
    
    isDummy = 1

    def objectIds(self, foo):
        return [self.title]

    def __repr__(self):
        return `self.title`

dummy = Dummy()

class EMarket(OFS.Folder.Folder):

    """An EMarket class.

    Provide the functions needed to contain an online shopping area.
    """

    # Specify a name for the item type:
    meta_type='EMarket'
    id       ='emarket'
    title    ='EMarket Object'
    icon     ='misc_/EMarket/emarket'
    RegistrationSubject = 'Registration Information'
    RegistrationFrom = 'manager@' + defaulthost              
    EMarketURL = ''                                          # where is 'home'?
    default_html = ''                                        # if we don't want index_html as the default doc.
    administratorEmail = RegistrationFrom                    
    requireLogin = 0
    addToCartText = 'Add Item To My Shopping Cart'
    shoppingCartTitle = 'My Shopping Cart'    
    allowShoppersToSell = 0
    biddingAllowed = 0
    refreshEnabled = 1
    defaultToAuction = 0
    processURL = ''

    _properties=({'id':'title', 'type': 'string', 'mode':'w'},
                 {'id':'RegistrationSubject',
                  'type':'string',
                  'mode':'w'},
                 {'id':'RegistrationFrom',
                  'type':'string',
                  'mode':'w'},
                 {'id':'EMarketURL',
                  'type':'string',
                  'mode':'w'},
                 {'id':'default_html',
                  'type':'string',
                  'mode':'w'},
                 {'id':'administratorEmail',
                  'type':'string',
                  'mode':'w'},
                 {'id':'requireLogin',
                  'type':'int',
                  'mode':'w'},
                 {'id':'addToCartText',
                  'type':'string',
                  'mode':'w'},
                 {'id':'shoppingCartTitle',
                  'type':'string',
                  'mode':'w'},
                 {'id':'allowShoppersToSell',
                  'type':'int',
                  'mode':'w'},
                 {'id':'biddingAllowed',
                  'type':'int',
                  'mode':'w'},
                 {'id':'refreshEnabled',
                  'type':'int',
                  'mode':'w'},
                 {'id':'defaultToAuction',
                  'type':'int',
                  'mode':'w'},
                 {'id':'randWords',
                  'type':'tokens',
                  'mode':'w'},
                 {'id':'processURL',
                  'type':'string',
                  'mode':'w'},
                 )

    manage_options = OFS.Folder.Folder.manage_options + (
        {'label':'DefaultDTML', 'action':'refresh_warning'},)

    # Specify how individual operations add up to "permissions":
    __ac_permissions__=(
	('View management screens', ('manage_tabs','manage_main')),
	('Change permissions',      ('manage_access',)           ),
	('Change EMarkets',         ('refresh_warning', 'refresh_dtml')),
	('View',                    ('',
                                     'showRequest',
                                     )
         ),
	)

    refresh_warning = HTMLFile('refresh_warning', globals())

    def refresh_dtml(self, REQUEST=None, ignoreDropErrors = 1):
        """ Refresh all the default DTML documents. """

        if not self.refreshEnabled:
            raise "Bad Request", "<html><br><br><center><b>Sorry.... no refresh allowed.</b></center></html>"
        
        errText = []
        
        errText.append(updateDocsInFolder(self, mainMarketDTML, REQUEST=REQUEST))
        errText.append(updateDocsInFolder(self.marketFolder, marketItemDTML, REQUEST=REQUEST))
        errText.append(updateDocsInFolder(self.shopperFolder, shopperDTML, REQUEST=REQUEST))
        if self.wampumbox:
            errText.append(updateDocsInFolder(self.wampumbox, wampumboxDTML, REQUEST=REQUEST))

        errs = string.join(errText)

        result = "<pre>We've Got Results...:\n" + errs + "</pre>"

        if REQUEST is not None:
            return MessageDialog(
                title='OK!',
                message=result + '<strong>%s</strong> Has default DTML docs now.' % self.id,
                action ='./manage_main',
                )

        return result

    def set_cookie(self, RESPONSE = None,
                   REQUEST = None, path = None, parents = None,
                   clientCookie = None, oldCookie = None):
        """
        Check the browser cookie...
        """
        #
        if RESPONSE is not None:
            if (path is not None) and (parents is not None):
                cookiePath = self.constructCookieURL(REQUEST, parents, path)
            else:
                cookiePath = '/'
                
            RESPONSE.setCookie('clientCookie', clientCookie, path = cookiePath)

        if oldCookie is not None:
            # we need to copy the temp shopper here and then delete the
            # temporary shopper info....
            # replace the old basket with the new...

            oldShopper = getattr(self.shopperFolder, oldCookie)
            newShopper = getattr(self.shopperFolder, clientCookie)
            newShopper.basket.items = oldShopper.basket.items
            newShopper.basket.UpdateShopper(newShopper)
            self.shopperFolder._delObject(oldCookie)

        return "clientCookie = '%s', path = '%s'" % (clientCookie, cookiePath)

    def checkEMail(self, email):
        """
        check to see if there is an email address among the shoppers.....
        """
        #
        found = 0

        for shopperID in self.shopperFolder.objectIds(['Shopper']):
            shopper = getattr(self.shopperFolder, shopperID)
            if shopper.email == email:
                found = 1
                break

        return found

    def FixupClassProperties(self):
        """ make sure class propteries are included in instance.... this is useful if you
            need to change the properties of a live instance
        """
        
        testClass = self.__class__()
        
        for p in testClass._properties:
            n = p['id']
            if not self.hasProperty(n):
                self._properties=self._properties + ({'id':n,
                                                      'type': p['type'],
                                                      'mode': p.get('mode','wd')},)
        return "OK!"

    def fixupFolder(self):
        """ fix a folder... utility for development.... dummy function"""

        #
        # find the transaction folder..
        #

        result = "Hmmm. not found"
        
        if hasattr(self, 'transactions'):
            self.transactionFolder = getattr(self, 'transactions')
            result = "OK!"

        return result
    
    def findEMail(self, email):
        """
        Find an email address among the shoppers.....
        """
        #
        foundInfo = ''

        for shopperID in self.shopperFolder.objectIds(['Shopper']):
            shopper = getattr(self.shopperFolder, shopperID)
            if shopper.email == email:
                foundInfo = foundInfo + "Shopper-ID : %s, password : %s " % (shopperID, shopper.passwd)
                break

        return foundInfo

    def findShopper(self, shopperID):
        """ Find information about a shopper based on 'id'. """

        if hasattr(self.shopperFolder, shopperID):
            shopper = getattr(self.shopperFolder, shopperID)
            foundInfo = "Shopper-ID : %s, password : %s " % (shopperID, shopper.passwd)
        else:
            foundInfo = "Can't find shopper: %s." % shopperID            

        return foundInfo

    def assignNewCookie(self, REQUEST=None, RESPONSE=None, PATH_INFO=None, PARENTS=None):
        """
        This method allows for assignment of 'random' cookies....
        so folks don't have to fill in forms just to shop around...
        """
        #
        errs = StringIO.StringIO()
        id = `time.time()`

        try:
            if RESPONSE is not None:
                result = Shopper.addShopper(self.shopperFolder, id = id, email='', name='',
                                            passwd = '', passwd2 = '', address = '', city = '',
                                            phone='', fax='',
                                            state = '', country = '', zip = '', passExcept = 1,
                                            noCheck = 1)

                self.set_cookie( REQUEST=REQUEST, RESPONSE=RESPONSE, path=PATH_INFO, parents=PARENTS, clientCookie= id)
                
                return "Successfully created user and set cookie...."
            else:
                raise RuntimeError, "Hello? This needs a RESPONSE!"
        except:
            traceback.print_exc(file=errs)
            errs.seek(0)
            return "failed to create new shopper. \n" + errs.read()
        
    def addShopper(self, id, email, name, passwd, passwd2, address, city, state,
                   phone, fax, country, zip, company='' ):
        """  addShopper creates a new shopper using the given cookie.... """

        ShopperError = Shopper.ShopperError
    
        result = Shopper.addShopper(self.shopperFolder, id, email, name,
                                    passwd, passwd2, address, city,
                                    state, phone, fax, country, zip, passExcept = 1,
                                    company=company)

        return result

    def addNewShopper(self, REQUEST = None, RESPONSE = None, PATH_INFO = None, PARENTS = None):
        """
        addNewShopper handles a special situation. We have a temporary user with
        no passwd, name etc. etc. etc.... We need to extract the name etc from
        the REQUEST, but *not* aqcuire the bogus values from the temporary user.
        This means we need to explicity get the from the REQUEST object.
        """
        #

        ShopperError = Shopper.ShopperError

        checkList = ['id','email','name','address',
                     'city','state','country','zip','phone','fax', 'company']

        missingList = []

        for item in checkList:
            if not REQUEST.has_key(item): 
                missingList.append(item)

        if len(missingList) > 0:
            raise "Bad Request", "Missing Data..."

        clientCookie = REQUEST.cookies.get('clientCookie',None)
        origURL = REQUEST.get('origURL',None)
        validateUser = REQUEST.get('validateUser',0)
        
        id = REQUEST['id']
        email = REQUEST['email']
        name = REQUEST['name']
        address = REQUEST['address']
        city = REQUEST['city']
        state = REQUEST['state']
        country = REQUEST['country']
        zip = REQUEST['zip']
        phone = REQUEST['phone']
        fax = REQUEST['fax']
        company = REQUEST['company']

        if not hasattr(self,'randWords'):
            self.randWords = randWords
            self.FixupClassProperties()

        randPass = random.choice(self.randWords) + random.choice(randSymbs) + random.choice(self.randWords)

        passwd = REQUEST.get('passwd',randPass)
        passwd2 = REQUEST.get('passwd2',randPass)

        if passwd == '':
            passwd = passwd2 = randPass
        
        try:
            msg = self.addShopper(id, email, name, passwd, passwd2, address, city,
                                                state, phone, fax, country, zip, company=company)

            if clientCookie is not None and hasattr(self.shopperFolder, clientCookie):
                # we need to copy the temp shopper here and then delete the
                # temporary shopper info....

                oldShopper = getattr(self.shopperFolder, clientCookie)
                newShopper = getattr(self.shopperFolder, id)
                # replace the old basket with the new...
                newShopper.basket.items = oldShopper.basket.items
                newShopper.basket.UpdateShopper(newShopper)
                self.shopperFolder._delObject(clientCookie)

                if origURL is not None:
                    origURL = string.replace(origURL, clientCookie, id)

            if RESPONSE is not None:
                #self.set_cookie(REQUEST = REQUEST, RESPONSE = RESPONSE, path=PATH_INFO, parents=PARENTS, clientCookie=id)
                result = "Successfully added new Shopper."

        except ShopperError, msg:
            result = "Sorry... %s" % msg

        return result


    def constructAbsURL(self, meta_type, parents, path, base, rel):
        """

            Construct an Absolute URL based on:

            meta_type: the metatype of the object, higher in the object tree
               relative to which 'rel' is based

            parents: the PARENTS of the current object

            path: the PATH_INFO of the current object

            base: the BASE0 from Zope

            rel: the path, relative to the object with the given meta_type.
            
        """
        # put a comment here to make python-mode happy.....
        
        foundIndex = -1  # nothing found yet...
        
        for index in range(len(parents)):
            item = parents[index]

            if item.meta_type == meta_type:
                foundIndex = index
                break
        
        while path[:1]=='/': path=path[1:]
        while path[-1:]=='/': path=path[:-1]
        path=string.split(path,'/')[:-1]

        if foundIndex == 0:
            newPath = path
        elif foundIndex == -1:
            raise EMarketError  # there was no object with the given meta_type.
        else:
            newPath=path[:-foundIndex]

        if base:
            return base+'/'+string.join(newPath,'/') + '/' + rel
        else:
            return string.join(newPath,'/') + '/' + rel

    def constructCookieURL(self, REQUEST, parents, path):
        """ construct cookie-root URL """
        
        foundIndex = -1  # nothing found yet...
        
        for index in range(len(parents)):
            item = parents[index]

            if item.meta_type == 'EMarket':
                foundIndex = index
                break
        
        if REQUEST.has_key('SCRIPT_NAME'):
            path = REQUEST['SCRIPT_NAME'] + path

        while path[0]=='/': path=path[1:]
        while path[-1]=='/': path=path[:-1]
        path=string.split(path,'/')[:-1]

        if foundIndex == 0:
            newPath = path
        elif foundIndex == -1:
            raise EMarketError  # there was no object with the given meta_type.
        else:
            newPath=path[:-foundIndex]

        return '/' + string.join(newPath,'/') 

    def shopperMarketItemAdd(self, REQUEST=None, RESPONSE=None, URL2=None):
        """
        Allow validated shoppers to add a Market Item

        'self' in this case is the folder to be added to.

        """
        #
        # check to make sure they are properly validated.....
        #

        title=REQUEST.get('title','')
        sku=REQUEST.get('sku',None)
        description=REQUEST.get('description','')
        imageTitle=REQUEST.get('imageTitle','')
        price=REQUEST.get('price',0.0)
        minPrice=REQUEST.get('minPrice',None)
        file=REQUEST.get('file',None)
        units=REQUEST.get('units','')
        quantityAvailable=REQUEST.get('quantityAvailable',None)
        duration=REQUEST.get('duration',None)
        category=REQUEST.get('category',None)
        deliveryInfo=REQUEST.get('deliveryInfo','')
        minQuantity=REQUEST.get('minQuantity',None)

        errs = StringIO.StringIO()
        
        if hasattr(self.shopperFolder,REQUEST['clientCookie']):
            theShopper = getattr(self.shopperFolder,REQUEST['clientCookie'])
            if not theShopper.validatedForSubmission:
                return "You must be validated for submission first..."
        else:
            RESPONSE.expireCookie('clientCookie')
            return "<center>There is no user with this Id.<br><br>\n" + \
                   "<a href=setUserCookie?validateUser=1&origURL=redirectToSalesItems>Enroll Here</a></center>"

        itemId = REQUEST['clientCookie'] + '-' + `time.time()`

        if quantityAvailable is None:
            quantityAvailable = 1
        elif type(quantityAvailable) == type(''):
            try:
                qa = string.strip(quantityAvailable)
                qa = string.replace(qa,',','')
                quantityAvailable = string.atoi(qa)
            except:
                quantityAvailable = 1

        if minQuantity is None:
            minQuantity = 1
        elif type(minQuantity) == type(''):
            try:
                qa = string.strip(minQuantity)
                qa = string.replace(qa,',','')
                minQuantity = string.atoi(qa)
            except:
                minQuantity = 1

        sku = itemId
        
        openTime = DateTime(time.time(), 'US/East-Indiana')
        
        if duration is not None:
            if type(duration) == type(''):
                try:
                    qa = string.strip(duration)
                    qa = string.replace(qa,',','')
                    duration = string.atof(qa)
                except:
                    duration = 7.0

            try:
                if duration < 3 or duration > 7:
                    duration = 7.0
                else:
                    duration = float(duration)
            except:
                duration = 7.0
        else:
            duration = 7.0

        closeTime = openTime + duration
        
        if category is not None and string.strip(category):
            theFolder = getattr(self.marketFolder, category)
        else:
            theFolder = self.marketFolder

        try:
            MarketItem.addMarketItem(theFolder,
                                     itemId, title=title, sku=sku, description=description,
                                     imageTitle=imageTitle, price=price, minPrice=minPrice,
                                     file=file, units=units, quantityAvailable=quantityAvailable,
                                     openTime=openTime, closeTime=closeTime, shopperID= REQUEST['clientCookie'],
                                     deliveryInfo=deliveryInfo, minQuantity=minQuantity, REQUEST=REQUEST)

            theNewItem = getattr(theFolder, itemId)

            if self.defaultToAuction:
                theNewItem.enableAuction = 1  # if defaultToAuction is true... make it an auction
                
            theShopper.salesItems.addItem( theNewItem, updateBooth=1)
            
        except:
            traceback.print_exc(file=errs)
            errs.seek(0)
            return MessageDialog(
                title = 'Sorry....',
                message = '</strong> Your Market item "%s" connot be entered.</strong>\n' % title + errs.read(),
                action = URL2
                )
            
        return MessageDialog(
            title = 'Entered.',
            message = '</strong> Your Market item "%s" has been entered.</strong>' % title,
            action = URL2
            )

    def checkSKU(self, sku):
        """ check validity of an sku..."""
        if self.getMarketItem(sku) != None:
            raise 'Bad Request', (
                """ The sku <em>%s</em> is in use. """ % sku )

    def getMarketItem(self, sku, currentFolder=None):
        """ find the market item for a certain sku. """

        if currentFolder is None:
            currentFolder = self.marketFolder

        result = None

        for item in currentFolder.objectValues(['MarketItem']):
            if item.sku == sku:
                result = item
                break

        if result is None:
            for folder in currentFolder.objectValues(['Folder']):
                result = self.getMarketItem( sku, folder )
                if result is not None:
                    break

        return result

    def removeMarketItem(self, sku, currentFolder=None):
        """ find the market item for a certain sku. """

        if currentFolder is None:
            currentFolder = self.marketFolder

        result = None

        for item in currentFolder.objectValues(['MarketItem']):
            if item.sku == sku:
                currentFolder.manage_delObjects([item.id])
                result = 1

        if result is None:
            for folder in currentFolder.objectValues(['Folder']):
                result = self.removeMarketItem( sku, folder )
                if result is not None:
                    break

        return result

    def SimpleSearch(self, word=None, lowerWord=None, folder=None):
        """
        search for a 'word' in the description or title of Market Items ...

        word = possibly non lower case word...
        lowerWord = lower case word... or None

        if word and lowerWord are both None, all items are returned....
        
        """

        results = []

        if word is None:
            searchWord = lowerWord
        else:
            searchWord = string.lower(word)

        if folder is None:
            folder = self.marketFolder

        for item in folder.objectValues(['Folder','MarketItem']):

            if item.meta_type == 'Folder':
                results = results + self.SimpleSearch( lowerWord = searchWord, folder = item)
            else:
                if searchWord is None:
                    results.append(item)
                elif string.find(string.lower(item.title), searchWord) != -1:
                    results.append(item)
                elif string.find(string.lower(item.description), searchWord) != -1:
                    results.append(item)

        return results

    def getValidItems(self, listOfItems):
        """ return the set of 'valid' (available) items from the set passed in...."""

        result = []
        
        #RDM:  If isAvailable is true, then items should be excluded here only
        #if they aren't valid just in the context of this folder; isAvailable
        #will be used to generate lists of available items out of the Catalog, 
        #so if an item shouldn't appear in general, isAvailable should be false.
        #I think the remaining check below should probably be moved to
        #isAvailable as well, but I'm not quite sure how to do it.
        for item in listOfItems:
            if not item.isAvailable(): continue
            if item.shopperID:
              if not hasattr(self.shopperFolder,item.shopperID): continue
            result.append(item)

        return result


    def getCategories(self, theObjects):

        nonEmptyList = []

        for folder in theObjects:
            if len(self.getValidItems(folder.objectValues(['MarketItem']))) > 0:
                nonEmptyList.append(folder)

        nonEmptyList.sort(cmpFolders)

        return nonEmptyList

    def doProductLayout(self, theObjects, maxCols):

        #
        # we need to layout as many rowss as needed
        # in order to show all the non-empty categories
        # using no more than maxCols columns....
        #

        theObjects = self.getCategories(theObjects)

        numObjs = len(theObjects)

        numRows = (numObjs + maxCols - 1)/maxCols

        if numRows == 1:
            numCols = numObjs
        else:
            numCols = maxCols

        rows = []

        for i in range(numRows):
            cols = []
            rows.append(cols)
            for j in range(numCols):
                theIndex = i*numCols + j
                try:
                    cols.append(theObjects[theIndex])
                except:
                    cols.append(dummy)

        return rows

