__doc__='''MarketItem product implementation'''

__version__='$Revision: 1.28 $'[11:-2]

from Globals import HTMLFile, MessageDialog, Persistent, DateTime
from Products.ZCatalog.CatalogAwareness import CatalogAware
import OFS.Folder
import time, string

addMarketItemForm=HTMLFile('marketItemAdd',globals())

def addMarketItem(self, id, title='', sku=None, description='', imageTitle='', price=0.0, minPrice=None,
                  file=None, units='', quantityAvailable=None, openTime = None, closeTime = None,
                  shopperID = None, deliveryInfo = '', minQuantity=None, REQUEST=None):
    
    """Add an MarketItem to a folder

    The argument 'self' will be bound to a folder.

    """
    #
    if sku is None or sku=='':
        # we have no sku... generate one:
        sku = `int(time.time())`

    if id == '':
        id = sku

    if type(quantityAvailable)==type(''):
        quantityAvailable = string.atoi(string.strip(quantityAvailable))
        
    mi = MarketItem()

    self.checkSKU(sku)
    
    mi.id=id
    mi.title=title
    mi.description=description
    mi.sku=sku
    mi.units=units
    mi.image = ''
    mi.quantityAvailable = quantityAvailable
    mi.openTime = openTime
    mi.closeTime = closeTime
    mi.shopperID = shopperID  # the id of the seller...
    mi.deliveryInfo = deliveryInfo
    mi.buyers = {} # this is to keep a list of offers/bids for this item
    mi.SetRelUrl(self)  # save the relative URL of this marketItem
    mi.enableAuction = 0 # sell this item using 'auction logic'

    try:
        mi.price=string.atof(price)
    except:
        mi.price=0.0

    if minPrice is None:
        mi.minPrice = mi.price
    else:
        try:
            mi.minPrice=string.atof(minPrice)
        except ValueError:
            mi.minPrice=mi.price

    if minQuantity is None:
        mi.minQuantity = 0
    elif type(minQuantity) == type(''):
        try:
            mi.minQuantity=string.atoi(minQuantity)
        except ValueError:
            mi.minQuantity=0
    else:
        mi.minQuantity = minQuantity

    if file is not None:
        try:
            mi.image = OFS.Image.Image('image',imageTitle,file)
        except:
            pass

    self._setObject(id, mi)
    
    if REQUEST is not None:
        return self.manage_main(self,REQUEST, update_menu=1)

class MarketItem(CatalogAware,OFS.Folder.Folder):
    """A MarketItem """

    # Specify a name for the item type:
    meta_type='MarketItem'

    # Specify a relative URL for the icon used to display icons:
    icon='misc_/EMarket/marketItem'

    manage_contents=OFS.Folder.Folder.manage_main
    manage_main=HTMLFile('marketItemEdit', globals())
    manage_image=HTMLFile('marketItemImage', globals())

    # Specify definitions for tabs:
    manage_options=[
	{'label':'Attributes', 'action':'manage_main'},
        {'label':'ImageUpload','action':'manage_image'},
	]

    #
    # add in the options from the Folder class.... change
    # the 'action' for Contents to 'manage_contents'
    #
    
    for item in OFS.Folder.Folder.manage_options:

        if item['label'] == 'Contents':
            manage_options.append(item.copy())  # be sure to muck with a *copy*
            manage_options[-1]['action'] = 'manage_contents'
        else:
            manage_options.append(item)
             
    # Specify how individual operations add up to "permissions":
    __ac_permissions__=(
	('View management screens', ('manage_tabs','manage_main','manage_attributes')),
	('Change permissions',      ('manage_access',)           ),
	('Change MarketItems',      ('manage_edit', 'manage_editImage')),
	('View',                    ('','index_html', 'image')),
	)

    isMarketItem = 1
    id = 'marketItem'
    title = ''
    description = ''
    sku = ''
    price = ''
    minPrice = None
    minQuantity = None
    units = ''
    image = ''
    quantitySold = 0          # hom many units have already sold?
    quantityAvailable = None  # how many units are available?
    openTime = None           # when should this item be made available?
    closeTime = None          # when should this item be withdrawn?
    shopperID = None          # which 'shopper' is the 'seller' of this item?
    buyers = None             # list of current buyers/bidders...
    relURL = ''               # start with relURL not set....
    deliveryInfo = ''         # no delivery info to start with...
    enabled = 1               # sale is on....
    enableAuction = 0         # don't sell this as an auction item
    
    # Provide no "view" interface, use the acquired one:
    # index_html=HTMLFile('index', globals())
   
    def manage_editImage(self, file, title='', REQUEST=None):
        """
        Separate function to change image....
        """
        #
        failed = 1
        try:
            self.image = OFS.Image.Image('image',title,file)
            failed = 0
        except:
            pass

	if REQUEST is not None:
            if failed:
                return MessageDialog(
                    title='Failed',
                    message='<strong>%s</strong> has not been edited.' % self.title,
                    action ='./manage_main',
                    )
            else:
                return MessageDialog(
                    title='Edited',
                    message='<strong>%s</strong> has been edited.' % self.title,
                    action ='./manage_main',
                    )
                

    def SetRelUrl(self, theMarketFolder):
        """

        Set the relative URL of this MarketItem to theMarketItem folder...

        note.. this really only works when the MarketItem is referred to in
        it's folder container.....
        
        """

        relURL = self.id

        parentFolder = theMarketFolder

        while parentFolder.id != "marketItems":
            relURL = parentFolder.id + '/' + relURL
            parentFolder = parentFolder.aq_parent

        self.relURL = relURL

    def shopperMarketItemEdit(self, title='', sku=None, description='', imageTitle='', price=0.0,
                              minPrice = None, file=None, units='', quantityAvailable=None, duration = None,
                              category = None, deliveryInfo = '', minQuantity = None, enableAuction = 0,
                              REQUEST=None, RESPONSE=None, URL2=None):
        """
        Allow validated shoppers to add a Market Item

        'self' in this case is the folder to be added to.

        """
        #
        if REQUEST.has_key('clientCookie') and (REQUEST['clientCookie'] == self.shopperID):

            if REQUEST.has_key('resetOpen'):
                # we need to reset the openTime...
                #
                REQUEST.set('openTime',DateTime('est'))

            self.manage_edit(title, sku = self.sku, description = description,
                             price = price, minPrice = minPrice, units = units,
                             quantityAvailable = quantityAvailable, deliveryInfo = deliveryInfo,
                             minQuantity = minQuantity, enableAuction = enableAuction, REQUEST = REQUEST)

            if file is not None:
                self.manage_editImage( file = file, title = imageTitle)

            if REQUEST is not None:
                return MessageDialog(
                    title='Edited',
                    message='<strong>%s</strong> has been edited.' % self.title,
                    action = 'redirectToSalesItems',
                    )
        else:
            return MessageDialog(
                title = 'Not Owner',
                message = '<strong>Not Owner</strong>. You are not the owner of this item.',
                action = 'redirectToEMarket',
                )
        
    def manage_edit(self, title, sku, description, price, units,
                    quantityAvailable=None, minPrice=None, deliveryInfo='',
                    minQuantity=None, enableAuction = 0, REQUEST=None):
	"""
        Change some item properties
	"""

        enabled = REQUEST.get('enabled',0)

        if type(price) == type(''):
            price = string.strip(price)
            price = string.replace(price,',','')
            price = string.replace(price, '$','')
            try:
                price = string.atof(price)
            except:
                price = None

        if type(minPrice) == type(''):
            minPrice = string.strip(minPrice)
            minPrice = string.replace(minPrice,',','')
            minPrice = string.replace(minPrice,'$','')
            try:
                minPrice = string.atof(minPrice)
            except:
                minPrice = None

        if type(quantityAvailable)==type(''):
            qa = string.strip(quantityAvailable)
            qa = string.replace(qa,',','')
            try:
                quantityAvailable = string.atoi(qa)
            except ValueError:
                if qa == 'None':
                    quantityAvailable = None
                else:
                    raise

        if type(minQuantity)==type(''):
            qa = string.strip(minQuantity)
            qa = string.replace(qa,',','')
            try:
                minQuantity = string.atoi(qa)
            except ValueError:
                if qa == 'None':
                    minQuantity = None
                else:
                    raise

	self.title=title

        if self.sku != sku:
            self.checkSKU(sku)
            self.sku = sku

        self.description=description
        self.price=price
        self.minPrice=minPrice
        self.quantityAvailable=quantityAvailable
        self.minQuantity = minQuantity
        self.units=units
        self.deliveryInfo = deliveryInfo
        self.enabled = enabled
        self.enableAuction = enableAuction

        if REQUEST.has_key('openTime'):
            openTime = REQUEST.get('openTime')

            if type(openTime) == type(''):
                openTime = DateTime(openTime)

            duration = REQUEST.get('duration',None)

            if duration is not None:
                if type(duration) == type(''):
                    try:
                        qa = string.strip(duration)
                        qa = string.replace(qa,',','')
                        duration = string.atof(qa)
                    except:
                        duration = 7.0

                try:
                    if (duration > 0 and duration < 3) or (duration > 7):
                        duration = 7.0
                    else:
                        duration = float(duration)
                except:
                    duration = 7.0
            else:
                duration = 7.0

            if duration > 0.0:
                self.openTime = openTime
                self.closeTime = openTime + duration

        elif REQUEST.has_key('closeTime'):
            self.closeTime = DateTime(REQUEST.get('closeTime','est'))

        #RDM: Let the catalog know that the properties may have changed.
        self.reindex_object()
        
	if REQUEST is not None:
	    return MessageDialog(
		title='Edited',
		message='<strong>%s</strong> has been edited.' % self.id,
		action ='./manage_main',
		)

	    
    def cmpType(self, name, aType):
        """ compare the type of the named property to the type of cmpType """
        return type(getattr(self, name)) == type(aType)

    def getBuyers(self, outbid = None, inbid=None):
        """
          get a list of dicts to describe the buyers of an item....

          if outbid is not None, get only buyers who are outbid at new bid level...
          if inbid is not None, get only buyers who are now in the 'running.'

          Each dict must have keys:
            'sku'       : the sku of the market item
            'shopperID' : the shopper id of the offer maker...
            'quantity'  : the quantity of the offer
            'offer'     : the amount of the offer

            reverse sort the list by offer amount.
            
        """
        
        result = []
        
        for shopperID in self.buyers.keys():
            b = {}

            if hasattr(self.shopperFolder, shopperID):

                theShopper = getattr(self.shopperFolder, shopperID)
                basketKey = self.id + '()'

                if theShopper.basket.items.has_key(basketKey):

                    theItem = theShopper.basket.items[ basketKey ]

                    result.append(b) # add this shopper to the list...

                    b['sku'] = self.sku
                    b['shopperID'] = shopperID
                    b['quantity'] = theItem.quantity
                    b['offer'] = theItem.offer


        result.sort(lambda x, y: cmp(x['offer'], y['offer']))
        result.reverse()

        #
        # result is now a sorted list of buyers sorted from highest offer
        # to lowest offer.
        #

        qa = self.quantityAvailable

        if outbid is not None:
            #
            # trim the list to include only those that are now
            # out-bid by the latest change...
            #
            
            while qa > 0:
                qa = qa - result[0]['quantity']
                del result[0]

        elif inbid is not None:
            #
            # trim the list to include only those that are now
            # still in the running 
            #
            newResult = []

            while qa > 0:
                newResult.append(result[0])
                qa = qa - result[0]['quantity']
                if qa < 0:
                    newQuantity = result[0]['quantity'] + qa
                    result[0]['reducedQuantity'] = newQuantity
                    
                del result[0]

            result = newResult

        return result

    def acceptOfferFromBuyer(self, REQUEST=None, shopperID=None):
        """ process the acceptance of an offer.


          return 1,msg,offer if all is OK.
          return 0,msg,None  if there is a problem....

        """
        #
        if REQUEST is not None:
            shopperID = REQUEST.get('shopperID','')

        result = 0
        msg = 'Never got the chance to check!'
        theItem = None

        if self.buyers.has_key(shopperID):

            bdict = self.buyers
            del bdict[shopperID]
            self.buyers = bdict

            if hasattr(self.shopperFolder, shopperID):

                theShopper = getattr(self.shopperFolder, shopperID)
                basketKey = self.id + '()'

                if theShopper.basket.items.has_key(basketKey):

                    theItem = theShopper.basket.items[ basketKey ]

                    if self.quantityAvailable is not None:
                        self.quantityAvailable = self.quantityAvailable - theItem.quantity
                        self.quantitySold = self.quantitySold + theItem.quantity

                    theShopper.basket.deleteItem( basketKey )
                    theShopper.basket.processItem( theItem )

                    msg = "OK!"
                    result = 1
                else:
                    msg = "This item is not in shopper's basket. "
            else:
                msg = "There is no shopper with this id."
        else:
            "There is no record that this shopper made an offer on this item."

        return result, msg, theItem

    def counterOfferFromBuyer(self, REQUEST=None, shopperID=None):
        """ process the acceptance of an offer.


          return 1,msg,offer if all is OK.
          return 0,msg,None  if there is a problem....

        """

        result = 0
        msg = 'Never got the chance to check!'
        theItem = None

        if REQUEST is not None:
            shopperID = REQUEST.get('shopperID','')
            offer = REQUEST.get('offer','')
            quantity = REQUEST.get('quantity','')

            if type(offer) == type(''):
                offer = string.strip(offer)
                offer = string.replace(offer,',','')
                offer = string.replace(offer,'$','')
                try:
                    offer = string.atof(offer)
                except:
                    raise "Bad Request", "Please enter a valid offer."

            if type(quantity) == type(''):
                quantity = string.strip(quantity)
                quantity = string.replace(quantity,',','')
                quantity = int(string.atoi(quantity))

        else:
            msg = "REQUEST was None"

        if self.buyers.has_key(shopperID):

            if hasattr(self.shopperFolder, shopperID):

                theShopper = getattr(self.shopperFolder, shopperID)
                basketKey = self.id + '()'

                if theShopper.basket.items.has_key(basketKey):

                    theItem = theShopper.basket.items[ basketKey ]
                    theItem.counterOffer = offer
                    theItem.counterQuantity = quantity
                    msg = "OK!"
                    result = 1
                else:
                    msg = "This item is not in shopper's basket. "
            else:
                msg = "There is no shopper with this id."
        else:
            msg = "There is no record that this shopper made an offer on this item.\n"
            msg = msg + "<!-- " + `self.buyers.keys()` + '-->\n'
            msg = msg + "<!-- " + `shopperID` + '-->\n'            

        return result, msg, theItem

    def checkOffer(self, shopper, offer, quantity, firmOffer=None):
        """ check to see if an offer is valid...

            returns  (errorStatus, statusCode)

            where:  errorStatus = 0 OK, 1 fatal error

                    statusCodes
                       1  attempt to drop quantity
                       2  not enough product remaining
                       3  attempt to drop offer
                       4  offer is below minPrice, or it's an auction
            

        """

        NO_ERROR, FATAL_ERROR = 0,1
        NO_STATUS, DROP_QUANTITY, NO_PRODUCT, DROP_OFFER, BELOW_MIN_OR_AUCTION = range(5)
        
        fatalErr = NO_ERROR
        statusCode = NO_STATUS
        basketKey = self.id + '()'
        
        if self.buyers.has_key(shopper.id) and (shopper.basket.items.has_key(basketKey) or
                                                shopper.pendingBasket.items.has_key(basketKey)):
            #
            #  This is an updated offer....
            #
            if shopper.basket.items.has_key(basketKey):
                item = shopper.basket.items[basketKey]
            else:
                item = shopper.pendingBasket.items[basketKey]

            if item.counterQuantity is not None:  # quantity can drop if it is countered with lower..
                oldQuantity = min([item.quantity,item.counterQuantity])
            else:
                oldQuantity = item.quantity

            if item.counterOffer is not None:     # price can drop if it is countered with lower...
                oldPrice = min([item.offer, item.counterOffer])
            else:
                oldPrice = item.offer

            if (quantity >= oldQuantity) and \
               (quantity <= self.quantityAvailable) and \
               (offer >= oldPrice):
                pass
            elif firmOffer is None:
                pass
            else:
                if quantity < oldQuantity:
                    statusCode = DROP_QUANTITY

                if quantity > self.quantityAvailable:
                    statusCode = NO_PRODUCT
                    
                if offer < oldPrice:
                    statusCode = DROP_OFFER

                fatalErr = FATAL_ERROR

        else:
            
            if quantity <= self.quantityAvailable:
                if offer < self.minPrice or self.enableAuction:
                    statusCode = BELOW_MIN_OR_AUCTION
            else:
                statusCode = NO_PRODUCT
                fatalErr = FATAL_ERROR
                
        if not fatalErr:
            if firmOffer is not None:
                b = self.buyers
                b[shopper.id] = 1
                self.buyers = b

        return fatalErr, statusCode

    def calcPrice(self, optionDict, theBasketItem):
        """ Calculate the price of a single item with the given optionDict....

        This is admittedly a kludge... but we needed something quick and easy.
        The option dict is a dict of optional features of MarketItems. These
        features might affect the price. The "price" attribute of a MarketItem
        is a 'base' price. If there is an optional feature, say color we add a
        property to the market item called 'colors' and list the possibility.
        If this property affects the price """

        #
        if theBasketItem.offer is not None:
            return theBasketItem.offer
        else:
            return self.price  # just return the base price for now....

    def offersAccepted(self):
        """ Are offers accepted? """
        return self.minPrice is not None and self.minPrice < self.price

    def PrincipiaSearchSource(self):
        """ This is the default index for searching for items. """
        return self.title+' '+self.description

    def isAvailable(self):
        """ Is this item avaialble for purchase?  If this returns false,
            the item should not appear in any lists of items presented
            to the customer.  This means catalog search need to check this."""
        if self.shopperID:
            if self.expired(): return 0
        if not self.enabled: return 0
        if (self.quantityAvailable<>None) and self.quantityAvailable<=0:
          return 0
        return 1

    def expired(self):
        """ is this item expired? """

        isExpired = 0
        
        if self.offersAccepted() or self.enableAuction:
            isExpired = DateTime().greaterThan(self.closeTime)

        return isExpired
