__doc__='''Shopper product implementation'''

__version__='$Revision: 1.42 $'[11:-2]

ShopperError = "Shopper Error."

from Globals import HTMLFile, MessageDialog, Persistent
import OFS.SimpleItem, Acquisition, AccessControl.Role
from Basket import Basket, BasketItem
from SalesBooth import SalesBooth
import string
import time

addShopperForm=HTMLFile('shopperAdd',globals())

def addShopper(self, id, email, name, passwd, passwd2, address, city, state,
               phone, fax, country, zip, REQUEST=None, company = '', passExcept = 0, noCheck = 0):
    """Add an shopper to a folder

    The argument 'self' will be bound to a folder.

    The arguments, 'id', 'passwd', and 'passwd2' just get us started...
    """
    #

    result = ''
    
    try:
        if not noCheck:
            if passwd != passwd2:
                raise ShopperError, "Passwords must match."

            if len(passwd) < 5:
                raise ShopperError, "Password must be longer than 4 characters..."

            if '@' not in email:
                raise ShopperError, "Email must contain an '@'."

            if '-' in id:
                raise ShopperError, "ID cannot contain a '-' character."

        i = Shopper(id, email, name, passwd, address, city, state,
                    phone, fax, country, zip, company)

        self._setObject(id, i)

        if not noCheck:
            i.validatedForSubmission = 1

        if REQUEST is not None:
            result = self.manage_main(self,REQUEST, update_menu=1)

    except ShopperError, msg:
        result = "Sorry.... %s" % msg

        if passExcept:
            raise ShopperError, msg  # re-raise the exception....

    return result

class Shopper(
    OFS.SimpleItem.Item,
    Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    ):

    """ A Shopper """

    # Specify a name for the item type:
    meta_type='Shopper'

    # Specify a relative URL for the icon used to display icons:
    icon='misc_/EMarket/shopper'

    # Specify definitions for tabs:
    manage_options=(
	{'label':'Properties', 'action':'manage_main'},
        {'label':'Shopping Basket', 'action':'shoppingCart'},
        {'label':'Login as Shopper','action':'loginAsUser'},
	{'label':'View',       'action':''},
	{'label':'Security',   'action':'manage_access'},
	)
 
    # Specify how individual operations add up to "permissions":
    __ac_permissions__=(
	('View management screens', ('manage_tabs','manage_main')),
	('Change permissions',      ('manage_access',)           ),
	('Change Shoppers',         ('manage_edit','loginAsUser')),
	('View',                    ('','manage_self_edit','manage_changePass')      ),
	)

    # Setup default class values....
    
    id = 'default'
    email = ''
    name = ''
    passwd = ''
    address = ''
    city = ''
    state = ''
    phone = ''
    fax = ''
    country = ''
    zip = ''
    company = ''
    validatedForSubmission = 0
    basket = None
    salesItems = None
    pendingBasket = None
    isShopper = 1
    transNum = 1
    transID = id + '-' + `transNum` + '-' + 'DEFAULT'
    transDict = {}
    addTime = time.time()

    def __init__(self, id, email, name, passwd, address='', city='', state='',
                 phone='', fax='', country='', zip='', company='' ):
        """
        Create a new shopper instance.
        """
        #
	self.id=id
        self.email = email
        self.name = name
        self.passwd = passwd
        self.address = address
        self.city = city
        self.state = state
        self.phone = phone
        self.fax = fax
        self.zip = zip
        self.company = company
        self.country = country
        self.transNum = 1
        self.transDict = {}
        self.basket = Basket(self)              # firm offers...
        self.pendingBasket = Basket(self)       # not yet firm offers....
        self.pendingBasket.ClearTransaction()
        self.salesItems = SalesBooth(self)
        self.transID = self.createTransID()
        self.addTime = time.time()

    # Provide interface for changing properties:

    manage_main=HTMLFile('shopperEdit', globals())

    def fixup(self):
        """ generic fixup function..."""
        if self.pendingBasket is None:
            self.pendingBasket = Basket(self)       # not yet firm offers....
            
        self.pendingBasket.ClearTransaction()
        return "OK!"

    def createTransID(self, makeNew=0, transNum = None):
        """  create a new (or retrieve old) transaction ID for a shopper. """

        if makeNew:
            self.transNum = self.transNum + 1  # this needs to be a new transaction...
        elif transNum is None:
            transNum = self.transNum
        else:
            self.transNum = transNum
            
        if self.transDict.has_key(self.transNum):
            transID = self.transDict[self.transNum]
        else:
            transID = self.id + '-' + `self.transNum` + '-' + `time.time()`
            td = self.transDict
            td[self.transNum] = transID
            self.transDict = td
            self.transID = transID
            
            self.basket.ClearTransaction()         # line up on new transaction....
            self.pendingBasket.ClearTransaction()  # clear out pending offers as well...
            
        return transID
    
    def parseOptions(self, optionDictString):
        """ convert an optionDict string into a tuple """
        return ""
        
    def addToBasket(self, REQUEST=None, RESPONSE=None):
        """ Add and item to the basket... get everything from the 'REQUEST'

        returns  statusMsg, statusCode

        where

            statusMsg = "OK-Low"  if all is well, but offer is too low...
                        "OK" if all is well and offer is high enough
                        "CONFIRM" if offer still needs to be confirmed
                        "Failed" if there is some unrecoverable problem

            statusCode =
                       -1 cannot find any MarketItem to match request
                       1  attempt to drop quantity
                       2  not enough product remaining
                       3  attempt to drop offer
                       4  offer is below minPrice, or it's an auction
        """

        if REQUEST is None:
            return "Hello? not much to do......"
        
        quantity = REQUEST.get('quantity',1)
        firmOffer = REQUEST.get('firmOffer',None)
        fromPending = REQUEST.get('fromPending',None)  # are we changing from pending?
        itemID = REQUEST.get('itemID',None)
        sku = REQUEST.get('sku',None)
        updateBasket=REQUEST.get('updateBasket',None)
        addApproved = 0  # add is approved and ready...
        
        if REQUEST.has_key('offer'):
            #
            # this is a bid.. not a straight purchase...
            #
            offer = REQUEST['offer']
        else:
            offer = None  # just a straight purchase..
            
        if type(quantity) == type(''):
            quantity = string.strip(quantity)
            quantity = string.replace(quantity,',','')
            quantity = int(string.atoi(quantity))

        if type(offer) == type(''):
            offer = string.strip(offer)
            offer = string.replace(offer,',','')
            offer = string.replace(offer,'$','')
            try:
                offer = string.atof(offer)
            except:
                offer = None

        optionDict={}

        if REQUEST.has_key('optionDict'):
            optionDict = REQUEST['optionDict']

        elif REQUEST.has_key('options'):
            options = string.split(string.strip(REQUEST['options'])[1:-1],',')
            for option in options:
                option = string.strip(option)[1:-1]  # get rid of quotes!
                optionDict[option] = REQUEST.get(option,'')

        item = self.getMarketItem(sku)

        if item is not None: # we found the item....

            if not item.enabled:
                raise "Bad Request", "Sorry this item is not available."
                
            if offer is not None and self.biddingAllowed:
                result, statusCode = item.checkOffer(self, offer, quantity, firmOffer)
                if not result and (firmOffer is not None):
                    addApproved = 1
                    
            else:
                addApproved, result, statusCode = 1,0,0

            if addApproved:

                returnCode = "OK"

                self.basket.addItem( anItem=item, quantity=quantity,
                                     optionDict = optionDict, updateBasket=updateBasket, offer=offer )

                #RDM If q<=0, addItem deleted it, and we're done.
                if quantity<=0: return "OK", statusCode

                basketItem = self.basket.getItem( anItem = item, optionDict = optionDict)

                if offer is not None and self.biddingAllowed:
                    if fromPending is not None:
                        #
                        # get offer from "pending basket" and put it in the 'real basket'.
                        #
                        self.pendingBasket.addItem(anItem=item, quantity=0, optionDict = optionDict,
                                                   updateBasket=updateBasket, offer=offer )

                    if basketItem.counterOffer is not None:
                        minPrice = basketItem.counterOffer
                    else:
                        minPrice = item.minPrice

                    if basketItem.counterQuantity is not None:
                        minQuantity = basketItem.counterQuantity
                    else:
                        minQuantity = item.minQuantity
                        
                    if offer >= minPrice:
                        returnCode = "OK"
                    else:
                        returnCode = "OK-Low"
                
                return returnCode, statusCode
            
            elif not result:
                if self.pendingBasket is None:
                    self.pendingBasket = Basket(self)
                    self.pendingBasket.ClearTransaction()
                    
                self.pendingBasket.addItem( anItem=item, quantity=quantity,
                                            optionDict = optionDict, updateBasket=updateBasket, offer=offer )

                return "CONFIRM", statusCode
            else:
                return "Failed", statusCode
        else:
            return "Failed", -1

    def clearPastTransactions(self):
        """ Clear out the past transactions of a shopper.."""
        self.basket.clearProcessedItems()
        return "OK!"
    
    def addSalesItem(self, REQUEST=None):
        """ Add and item for sale... get everything from the 'REQUEST' """

        if REQUEST is None:
            return "Hello? not much to do......"
        
        quantityAvailable = REQUEST.get('quantity',1)
        itemID = REQUEST.get('itemID')
        sku = REQUEST.get('sku')
        updateBooth=REQUEST.get('updateBooth',None)
        enabled=REQUEST.get('enabled',0)

        price = REQUEST.get('price')

        if type(price) == type(''):
            price = string.strip(price)
            price = string.replace(price,',','')
            price = string.replace(price, '$','')
            try:
                price = string.atof(price)
            except:
                price = None

        minPrice = REQUEST.get('minPrice')

        if type(minPrice) == type(''):
            minPrice = string.strip(minPrice)
            minPrice = string.replace(minPrice,',','')
            minPrice = string.replace(minPrice,'$','')
            try:
                minPrice = string.atof(minPrice)
            except:
                minPrice = None

        item = None

        item= self.getMarketItem(sku)

        if item is not None:
            item.quantityAvailable=quantityAvailable

            if price is not None:
                item.price = price
            else:
                item.price = 0.0

            if minPrice is not None:
                item.minPrice = minPrice
            else:
                item.minPrice = 0.0

            item.enabled = enabled
            
            self.salesItems.addItem( anItem=item, updateBooth=updateBooth )
            return "OK","No problem..."
        else:
            return "Failed","Hmm can't find that market item..."

    def getSalesItems(self):
        """ return the active sales items for a seller...."""

        theItems = self.salesItems.items
        result = []

        for k,v in theItems.items():
            if self.getMarketItem(v.sku) == None:
                del theItems[k]  # get rid of items that have been deleted....
            else:
                result.append(v)
        return result

    def loginAsUser(self, REQUEST=None, RESPONSE=None, PARENTS=None, PATH_INFO=None):
        """ set the browser cookie to user id for this user..."""

        result = self.set_cookie(RESPONSE = RESPONSE, REQUEST=REQUEST,
                                 path=PATH_INFO, parents=PARENTS, clientCookie=self.id)

        if REQUEST is not None:
            result = MessageDialog(
                title='Logged In',
                message='You are logged in as <strong>%s</strong><br> %s.' % (self.id, result),
                action = "./manage_main")
        else:
            result = "OK"

        return result

    def setProcessed(self, transID):
        """ check transID for current-status..."""

        result = 0
        
        if transID == self.transID:
            result=1
            self.basket.processTransaction(transID)

        return result
    
    def manage_changePass(self, oldpass, passwd, passwd2, origURL=None, REQUEST=None, passExcept=0):
        """ Change the password. """

        try:

            result = "OK"

            if oldpass != self.passwd:
                raise ShopperError, "Bad password."

            if len(passwd) < 5:
                raise ShopperError, "Password must be longer than 4 characters..."

            if passwd != passwd2:
                raise ShopperError, "Passwords must match..."

            self.passwd = passwd

            if REQUEST is not None:
                if origURL is None:
                    origURL = './manage_main'

                result = MessageDialog(
                    title='Edited',
                    message='<strong>%s</strong>\'s password has been edited.' % self.id,
                    action = origURL
                    )

        except ShopperError, msg:
            result = "<html><body bgcolor=ffffff><center><h3>Sorry %s</h3></center></body></html>" % msg
            if passExcept:
                raise ShopperError, result

        return result

    def manage_self_edit(self, email, name, address, city, state, zip, country='', company='',
                         phone='', fax='', origURL=None, REQUEST=None):
        """
        Allow user to manage themselves.....

        """
        #

        if REQUEST.has_key('clientCookie') and (REQUEST['clientCookie'] == self.id):
            result = self.manage_edit(email, name, address, city, state, zip, country, phone=phone, fax=fax,
                                      company=company, origURL=origURL, REQUEST=REQUEST)

            if REQUEST.has_key('passwd') and REQUEST['passwd']:
                self.manage_changePass(oldpass = self.passwd,
                                       passwd = REQUEST.get('passwd',''),
                                       passwd2 = REQUEST.get('passwd2',''),
                                       passExcept = 1)
        else:
            result = "Sorry.... you are not logged in as this shopper...."

        return result
    
    def manage_edit(self, email, name, address, city, state, zip, country='', company='',
                    phone='', fax='', passwd='', passwd2='', origURL=None, REQUEST=None):
	"""
	Edit a shopper
	"""
        try:
            if '@' not in email:
                raise ShopperError, "Email must contain an '@'."
            else:
                if passwd == '':
                    passwd = self.passwd
                    passwd2 = self.passwd

                if len(passwd) < 5:
                    raise ShopperError, "Password must be longer than 4 characters..."

                if passwd != passwd2:
                    raise ShopperError, "Passwords must match..."

                self.passwd = passwd
                self.email = email
                self.name = name
                self.address = address
                self.city = city
                self.state = state
                self.zip = zip
                self.country = country
                self.company = company
                self.phone = phone
                self.fax = fax

                if REQUEST is not None:
                    if origURL is None:
                        origURL = './manage_main'

                    result = MessageDialog(
                        title='Edited',
                        message='<strong>%s</strong> has been edited.' % self.id,
                        action = origURL
                        )
                        
        except ShopperError, msg:
            result = "Sorry %s " % msg

        return result




